#!/usr/bin/env python3

# Copyright 2016 Christ Stefan <contact@stefanchrist.eu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Simple program to sort MP3 files into the folder hierarchy
#
#    <output dir>/<artist>/<album>/<track>.mp3
#
# I used it to sort thousand of MP3 files recoveried by 'photorec'
# (http://www.cgsecurity.org/wiki/PhotoRec). You have to install
# utility 'mp3info':
#
#    sudo apt-get install mp3info
#
# Usage:
#
#    mp3sort <input directory of files> <output directory prefix>
#
# then disable 'dryrun' in the source code. Sorry no commandline
# switch for it now. It's save to reexecute the script multiple
# times. It will not overwrite existing files.


import sys
import os
from subprocess import Popen, PIPE, DEVNULL


def getNewName(path):
    p = Popen([b"mp3info", b"-p", b"%a\n%l\n%t", path],
              stdout=PIPE,
              stderr=DEVNULL)
    e = p.wait()
    stdout = p.stdout.read()
    p.stdout.close()

    if e != 0:
        raise RuntimeError("Mp3info fails with error code %d" % (e,))

    artist, album, track = stdout.split(b"\n")
    if len(artist) == 0 or len(album) == 0 or len(track) == 0:
        raise ValueError("No valid track information in file!")

    # TODO Try to get encoding information form mp3 tag information and convert
    # to local system encoding. So you cannot use the byte buffer as a
    # filename, because it contains illegal characters.

    return os.path.join(artist, album, track + b".mp3")


def moveFile(path, new_path):
    os.makedirs(os.path.dirname(new_path), exist_ok=True)

    if os.path.exists(new_path):
        print("ERROR desitination file %s already exists!" % (new_path,),
              file=sys.stderr)
    else:
        os.rename(path, new_path)


def moveFileAndTryToSTrip(path, output_path, new_filename):
    new_path = os.path.join(output_path, new_filename)
    try:
        moveFile(path, new_path)
    except OSError as e:
        print("Error while moving file %s: %s" % (path, e),
              file=sys.stderr)
        # Mostly Error:
        #    [Errno 84] Invalid or incomplete multibyte or wide character: b'MP3/The Lord Of The Rings/Die R\xfcckkehr Des K\xf6nigs'
        # Try to fix it with a stripped version of the new filename.
        new_filename_stripped = bytes([c for c in new_filename if 0 < c < 128])
        print("Stripping filename %s -> %s" % (new_filename, new_filename_stripped),
              file=sys.stderr)
        new_path_stripped = os.path.join(output_path, new_filename_stripped)
        try:
            moveFile(path, new_path_stripped)
        except OSError as e:
            print("Error while moving file %s: %s" % (path, e), file=sys.stderr)

def main():
    if len(sys.argv) != 3:
        print("Usage: %s <input directory of files> <output directory prefix>" % (sys.argv[0],))
        return 1

    dryrun = True
    input_path = sys.argv[1].encode(sys.getdefaultencoding())
    output_path = sys.argv[2].encode(sys.getdefaultencoding())

    # Get all input files
    files = (os.path.join(input_path, filename)
             for filename in os.listdir(input_path))
    files_mp3 = (path for path in files
                 if os.path.isfile(path) and path.endswith(b".mp3"))

    # Move files
    for path in files_mp3:
        try:
            new_filename = getNewName(path)
        except Exception as e:
            print("ERROR in file %s: %s" % (path, e), file=sys.stderr)
            continue

        if new_filename is None:
            continue

        if dryrun:
            print("mv", path, new_path)
            continue

        moveFileAndTryToSTrip(path, output_path, new_filename)

    return 0


if __name__ == "__main__":
    sys.exit(main())
